@@ -43,12 +43,31 @@ credits = Credits('pyperplan',
                   'Pyperplan is a lightweight STRIPS planner written in Python.\nPlease note that Pyperplan deliberately prefers clean code over fast code. It is designed to be used as a teaching or prototyping tool. If you use it for paper experiments, please state clearly that Pyperplan does not offer state-of-the-art performance.\nIt was developed during the planning practical course at Albert-Ludwigs-Universität Freiburg during the winter term 2010/2011 and is published under the terms of the GNU General Public License 3 (GPLv3).\nPyperplan supports the following PDDL fragment: STRIPS without action costs.'
                 )
 
-class EngineImpl(unified_planning.engines.Engine,
-                 unified_planning.engines.mixins.OneshotPlannerMixin,
-                 unified_planning.engines.mixins.CompilerMixin):
-    def __init__(self, **options):
-        if len(options) > 0:
-            raise
+class EngineImpl(
+        unified_planning.engines.Engine,
+        unified_planning.engines.mixins.OneshotPlannerMixin,
+        unified_planning.engines.mixins.CompilerMixin
+    ):
+    """ Implementation of the up-pyperplan Engine. """
+
+
+    def __init__(self, search: str = "wastar", heuristic: Optional[str] = "hadd"):
+        unified_planning.engines.Engine.__init__(self)
+        up.engines.mixins.OneshotPlannerMixin.__init__(self)
+        up.engines.mixins.CompilerMixin.__init__(self)
+        if search not in SEARCHES:
+            raise up.exceptions.UPUsageError(f'{search} not supported!')
+        if heuristic not in HEURISTICS:
+            raise up.exceptions.UPUsageError(f'{heuristic} not supported!')
+        self._search = SEARCHES[search]
+        self._heuristic = HEURISTICS[heuristic]
+        if search in ["bfs", "ids", "sat"]:
+            self._heuristic = None
+            self._optimal = True
+        elif search == "astar" and heuristic in ["hmax", "blind", "lmcut"]:
+            self._optimal = True
+        else:
+            self._optimal = False
 
     @property
     def name(self) -> str:
@@ -60,6 +79,7 @@ class EngineImpl(unified_planning.engines.Engine,
         supported_kind.set_problem_class('ACTION_BASED') # type: ignore
         supported_kind.set_typing('FLAT_TYPING') # type: ignore
         supported_kind.set_typing('HIERARCHICAL_TYPING') # type: ignore
+        supported_kind.set_quality_metrics("PLAN_LENGTH") # type: ignore
         return supported_kind
 
     @staticmethod
@@ -70,9 +90,16 @@ class EngineImpl(unified_planning.engines.Engine,
     def supports_compilation(compilation_kind: CompilationKind) -> bool:
         return compilation_kind == CompilationKind.GROUNDING
 
+    @staticmethod
+    def resulting_problem_kind(
+        problem_kind: ProblemKind,
+        compilation_kind: Optional[CompilationKind] = None,
+    ) -> ProblemKind:
+        return ProblemKind(problem_kind.features)
+
     @staticmethod
     def satisfies(optimality_guarantee: up.engines.OptimalityGuarantee) -> bool:
-        return False
+        return optimality_guarantee == up.engines.OptimalityGuarantee.SATISFICING
 
     @staticmethod
     def get_credits(**kwargs) -> Optional[unified_planning.engines.Credits]:
@@ -89,13 +116,15 @@ class EngineImpl(unified_planning.engines.Engine,
         return CompilerResult(grounded_problem, partial(up.engines.compilers.utils.lift_action_instance, map=rewrite_back_map), self.name, [])
 
     def _solve(self, problem: 'up.model.AbstractProblem',
-               callback: Optional[Callable[['up.engines.PlanGenerationResult'], None]] = None,
+               heuristic: Optional[Callable[["up.model.state.ROState"], Optional[float]]] = None,
                timeout: Optional[float] = None,
                output_stream: Optional[IO[str]] = None) -> 'up.engines.results.PlanGenerationResult':
         '''This function returns the PlanGenerationResult for the problem given in input.
         The planner used to retrieve the plan is "pyperplan" therefore only flat_typing
         is supported.'''
         assert isinstance(problem, up.model.Problem)
+        if heuristic is not None:
+            warnings.warn('Pyperplan does not support custom heuristic.', UserWarning)
         if timeout is not None:
             warnings.warn('Pyperplan does not support timeout.', UserWarning)
         if output_stream is not None:
@@ -103,18 +132,23 @@ class EngineImpl(unified_planning.engines.Engine,
         self.pyp_types = {}
         dom = self._convert_domain(problem)
         prob = self._convert_problem(dom, problem)
-        search = SEARCHES["bfs"]
         task = _ground(prob)
-        heuristic = None
-        # if not heuristic_class is None:
-        #     heuristic = heuristic_class(task)
-        solution = _search(task, search, heuristic)
+        h = None if self._heuristic is None else self._heuristic(task)
+        solution = _search(task, self._search, h)
         actions: List[up.plans.ActionInstance] = []
         if solution is None:
-            return up.engines.PlanGenerationResult(PlanGenerationResultStatus.UNSOLVABLE_PROVEN, None, self.name)
+            if self._search in ["bfs", "ids", "astar", "wastar", "gbf"]:
+                status = PlanGenerationResultStatus.UNSOLVABLE_PROVEN
+            else:
+                status = PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY
+            return up.engines.PlanGenerationResult(status, None, self.name)
         for action_string in solution:
             actions.append(self._convert_string_to_action_instance(action_string.name, problem))
-        return up.engines.PlanGenerationResult(PlanGenerationResultStatus.SOLVED_SATISFICING, up.plans.SequentialPlan(actions), self.name)
+        if self._optimal and len(problem.quality_metrics) > 0:
+            status = PlanGenerationResultStatus.SOLVED_OPTIMALLY
+        else:
+            status = PlanGenerationResultStatus.SOLVED_SATISFICING
+        return up.engines.PlanGenerationResult(status, up.plans.SequentialPlan(actions), self.name)
 
     def _convert_string_to_action_instance(self, string: str, problem: 'up.model.Problem') -> 'up.plans.ActionInstance':
         assert string[0] == "(" and string[-1] == ")"
@@ -238,3 +272,20 @@ class EngineImpl(unified_planning.engines.Engine,
         new_t = PyperplanType(type.name, father)
         self.pyp_types[type.name] = new_t
         return new_t
+
+
+class OptEngineImpl(EngineImpl):
+    def __init__(self, search: str = "astar", heuristic: Optional[str] = "lmcut"):
+        if search not in ["astar", "bfs", "ids"]:
+            raise up.exceptions.UPUsageError(f'{search} not supported!')
+        if heuristic not in ["hmax", "blind", "lmcut"]:
+            raise up.exceptions.UPUsageError(f'{heuristic} not supported!')
+        EngineImpl.__init__(self, search, heuristic)
+
+    @property
+    def name(self) -> str:
+        return "PyperplanOpt"
+
+    @staticmethod
+    def satisfies(optimality_guarantee: up.engines.OptimalityGuarantee) -> bool:
+        return True
